# pylint: disable=missing-docstring
#
# Load sample data included with GMT (downloaded from the GMT cache server).

from .tutorial import (
    load_japan_quakes,
    load_ocean_ridge_points,
    load_sample_bathymetry,
    load_usgs_quakes,
)
from .earth_relief import load_earth_relief
