# pylint: disable=missing-docstring
#
# Low-level wrapper for the GMT C API.
#
# The pygmt.clib.Session class wraps the GMT C shared library (libgmt) with a
# pythonic interface. Access to the C library is done through ctypes.

from .session import Session
